# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::Base.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Feature-Structures for perl: abstract base class
#
###############################################################

package Data::FStructure::Base;
# System modules
use Carp;
#require Exporter;
# 3rd party exstensions
# user extension modules
use Math::PartialOrder qw($TYPE_TOP);
#use Data::FStructure::perldata;
#use Set::Hashed;
@ISA       = qw();
#@EXPORT   = qw();
#@EXPORT_OK = (
#	     );
#%EXPORT_TAGS =
#  (
#  );

###############################################################
# Package-global variables
###############################################################
our $VERSION = 0.04;

###############################################################
# warnings
###############################################################
sub warn_abstract ($) {
  if ($^W && $VERBOSE) {
    my $warn =
      "Attempt to find non-existant method `$_[0]' via Data::FStructure::Base.\n"
      ." > (Did you forget to load a module or implement an abstract method?)";
    if ($VERBOSE > 1) { cluck($warn); }
    else { carp($warn); }
  }
}


###############################################################
# Constructor (ABSTRACT)
###############################################################
###############################################################
# Constructor
#   + Data::FStructure->new( {hierarchy=>$h} )
#   + Object-structure:
#     {
#      hierarchy       => Math::PartialOrder
#      Qset (nodes)    => { id1 => node1, ... } (literal hash)
#      q0 (root)       => Scalar node
#      qcounter        => Scalar integer (counts nodes)
#      master          => Scalar-ref: root of largest embedding FStructure
#      #theta (types)   => qd::fs::theta
#      #delta (ftrs)    => qd::fs::delta
#      #pathsep (opt)   => Scalar path-separator
#      #pathre (opt)    => Scalar pathsep-regexp
#     }
#   + Args:
#     { hierarchy => $h,
#       qcounter => $i,
#       root => $nodeid
#       theta => Data::FStructure::theta,
#       delta => Data::FStructure::delta,
#     }
###############################################################
sub new {
  my $proto = shift;
  my $args = shift;
  my $master = undef;
  my $self =
    {
     hierarchy => (defined($args) && defined($args->{hierarchy})
		   ? $args->{hierarchy}
		   : (ref($proto)
		      ? $proto->hierarchy
		      : undef)),
     #Qset => Set::Hashed->new(),
     Qset => {},
     q0 => undef,
     #theta => {},
     #delta => {},
     qcounter => 0,
     master => \$master,
    };

  # parse arguments & copy prototype's values
  # (only the following:)
  foreach (qw(hierarchy qcounter)) {
    if (exists($args->{$_})) {
      # arguments override prototype-values
      $self->{$_} = $args->{$_};
    }
    elsif (ref($proto) && exists($proto->{$_})) {
      $self->{$_} = $proto->{$_};
    }
  }

  # Bless into whatever package was given us...
  bless $self, ref($proto) || $proto;

  # set root if requested
  if (exists($args->{q0}))      { $self->q0($args->{q0}); }
  elsif (exists($args->{root})) { $self->q0($args->{root}); }

  # do subclass initialization
  $self->initialize($args) if ($self->can('initialize'));

  # return the new FS
  return $self;
}

###############################################################
# Default inititialization -- NONE
###############################################################

###############################################################
# Accessors/manipulators
###############################################################

# $h = $fs->hierarchy()
# $fs->hierarchy($h)
*_hierarchy = \&hierarchy;
sub hierarchy {
  return exists($_[1]) ? ($_[0]->{hierarchy} = $_[1]) : $_[0]->{hierarchy};
}

# $set = $fs->Qset();
# Qset($set)
*nodeset = \&Qset;
*_Q = \&Qset;
sub Qset { return exists($_[1]) ? ($_[0]->{Qset} = $_[1]) : $_[0]->{Qset}; }


# $q0id = $fs->q0()
# $q0id = $fs->q0($node)
*root = \&q0;
sub q0 {
  return $_[0]->{q0} unless (exists($_[1]));

  if (defined($_[1])) {
    # if we say it's the root, it'd better exist...
    $_[0]->add_node($_[1]);
    # if master was undefined, it shouldn't be now
    $_[0]->_master_q0($_[1]) unless (defined($_[0]->_master_q0));
  }
  else {
    # $fs->q0(undef) means remove the root...
    # are we removing the master root?
    $_[0]->_master_q0(undef) if
      (defined($_[0]->{q0}) &&
       defined($_[0]->_master_q0) &&
       $_[0]->{q0} eq $_[0]->_master_q0);
  }

  return $_[0]->{q0} = $_[1];
}


sub _theta { _warn_abstract('_theta'); }
sub _delta { _warn_abstract('_delta'); }


# $i = qcounter(), qcounter($i)
sub qcounter {
  return exists($_[1]) ? ($_[0]->{qcounter} = $_[1]) : $_[0]->{qcounter};
}


# $nodeid_ref = _master()
*_master_root_ref = \&_master;
sub _master { return $_[0]->{master}; }

# $nodeid = _master_q0
*_master_root_node = \&_master_q0;
sub _master_q0 {
  return exists($_[1]) ? (${$_[0]->{master}} = $_[1]) : ${$_[0]->{master}};
}

###############################################################
# Low-level utilities
###############################################################

# $new_nodeid = $fs->_gensym_node()
# $new_nodeid = $fs->_gensym_node($fs2)
# --> returns an id not yet present in $fs1 (or $fs2, if given)
sub _gensym_node {
  my ($i,$nodes1,$nodes2);
  $nodes1 = $_[0]->{Qset};
  if (defined($_[1])) {
    $i = $_[0]->{qcounter} > $_[1]->{qcounter}
         ? $_[0]->{qcounter}
         : $_[1]->{qcounter};
    $nodes2 = $_[1]->{Qset};
  } else {
    $i = $_[0]->{qcounter};
  }
  while (1) {
    last unless (exists($nodes1->{'q'.$i}) ||
		 (defined($nodes2) && exists($nodes2->{'q'.$i})));
    ++$i;
  }
  $_[0]->{qcounter} = $i + 1;
  return 'q'.$i;
}

# $rvalue = $fs->_get_rvalue($val)
#   + interprets tied references as substructures (w/ 'tied()')
#   ---> now uses gobble() interface! (old tied() hashes are broken!)
#sub _get_rvalue {
#  return undef unless (defined($_[1]));
#  return $_[1] unless (ref($_[1]));
#  return tied(%{$_[1]}) if
#    (ref($_[1]) eq 'HASH' &&
#     UNIVERSAL::isa(tied(%{$_[1]}), __PACKAGE__ ));
#  return tied(@$_[1]) if
#    (ref($_[1]) eq 'ARRAY' &&
#     UNIVERSAL::isa(tied(@{$_[1]}), __PACKAGE__ ));
#  #if ($INCORPORATE) { ... }
#  return $_[1];
#}


###############################################################
# Basic feature-structure manipulation
###############################################################
# @nodes = $fs->Q();
# @nodes = $fs->Q(@ids);
# ----> no addition, just lookup
*nodes = \&Q;
sub Q {
  return
    (exists($_[1])
     ? @{$_[0]->{Qset}}{@_[1..$#_]}
     : values(%{$_[0]->{Qset}}));
}


# $bool = $fs->is_empty()
# --> true if
#     1. no root node defined, OR
#     2. root node has no features
sub is_empty {
  return (!defined($_[0]->{root})
	  || !$_[0]->has_node($_[0]->{root})
	  || !$_[0]->features($_[0]->{root}));
}

# $bool = $fs->has_node($nodeid)
# $bool = $fs->has_nodes(@nodeids)
# ALIASES: has
sub has_node { return (defined($_[1]) && exists($_[0]->{Qset}{$_[1]})); }
*has = \&has_nodes;
sub has_nodes {
  grep {
    return '' unless (defined($_) && exists($_[0]->{Qset}{$_}));
  } @_[1..$#_];
  return 1;
}


# @nodeids = $fs->add_nodes(@nodes)
# ALIASES: add, ensure_nodes, ensure
*add = \&add_nodes;
*ensure_nodes = \&add_nodes;
*ensure = \&add_nodes;
sub add_nodes {
  my $fs = shift;
  return qw() unless (@_);
  my @ids = map { defined($_) ? $_ : $_[0]->_gensym_node() } @_;
  @{$_[0]->{Qset}}{@ids} = @_;
  return map { "$_" } @ids;
}
# $nodeid = $fs->add_node($node);
sub add_node {
  my $node = defined($_[1]) ? $_[1] : $_[0]->_gensym_node;
  return $_[0]->{Qset}{$node} = $node;
}

# $fs->remove_nodes(@nodeids)
# $fs->remove_node(@nodeids)
# $fs->remove(@nodeids)
*remove_node = \&remove_nodes;
*remove = \&remove_nodes;
sub remove_nodes {
  my $fs = shift;
  delete($fs->{Qset}{@_});

  $fs->{theta}->remove(@_);               # types
  $fs->{delta}->remove_transits_from(@_); # transistions from
  $fs->{delta}->remove_transits_to(@_);   # transistions to

  # check if we removed the root node (also handles master-root)...
  $fs->root(undef) unless (defined($fs->{q0}) && exists($fs->{Qset}{$fs->{q0}}));
  return $fs;
}

# $fs = $fs->remove_theta(@nodeids) : REQUIRED
# $fs = $fs->remove_delta(@nodeids) : REQUIRED


# $fs = $fs->rename_nodes($old => $new, ... )
# $fs = $fs->rename_node($old => $new, ... )
#### $fs = $fs->rename($old => $new, ... )
*rename_node = \&rename_nodes;
*replace = \&rename_nodes;
#*rename = \&rename_nodes;
sub rename_nodes {
  my ($fs,%rename) = @_;
  my ($q,$tmp);

  # rename actual nodes
  delete($fs->{Qset}{keys(%rename)});
  $fs->{Qset}{values(%rename)} = values(%rename);

  # alter type-map
  $fs->{theta}->replace(%rename);  # alter type-map
  $fs->{delta}->replace(%rename);  # alter feature-transitions

  # rename root node if requested
  if (defined($fs->{q0}) && exists($rename{$fs->{q0}})) {
    $fs->root($rename{$fs->{q0}});
  }
  # rename master root node if requested
  if (defined($fs->_master_root_node)
      && exists($rename{$fs->_master_root_node})) {
    $fs->_master_root_node($rename{$fs->_master_root_node});
  }
  return $fs;
}




# $node = $fs->nvalue(\@path), $fs->nvalue(\@path,$node)
sub delta_star {
  return $_[0]->_get_nvalue($_[1]) if (!exists($_[2]));
  return $_[0]->_set_nvalue(@_[1..2]);
}
*nvalue = \&delta_star;

sub _get_nvalue {
  my $fs = shift;
  my $path = shift;
  return $fs->{q0} unless ($path);
  my $node = $fs->{q0};
  my ($ftr);
  foreach $ftr (@$path) {
    return undef unless (defined($node = $fs->delta($node,$ftr)));
  }
  return $node;
}
# $nodeid = $fs->_get_or_create_nvalue(\@path)
sub _get_or_create_nvalue {
  my ($fs,$path) = @_;

  # is the path empty ?
  unless ($path && @$path && defined($path->[0])) {
    return $fs->{q0} if (defined($fs->{q0}));
    return $fs->{q0}($fs->_gensym_node);
  }

  # go get it...
  my ($node);
  my $last = $fs->root;
  foreach (@$path) {
    unless (defined($node = $fs->delta($last,$_))) {
      $fs->add_node($node = $fs->_gensym_node);
      $fs->delta($last,$_,$node);
    }
    $last = $node;
  }
  return $last;
}
sub _set_nvalue {
  my ($fs,$path,$dest) = @_;
  $fs->add_node($dest); # ensure we've got the node
  return $fs->root($dest) unless ($path && @$path);
  my $src = $fs->_get_or_create_nvalue([@$path[0..$#_path-1]]);
  $fs->delta($src,$path->[$#$path],$dest);
  return $dest;
}

# $fs2 = $fs1->substructure($node)
sub substructure {
  my ($fs,$node) = @_;

  return $fs if
    (defined($node) && defined($fs->{q0}) && $node eq $fs->{q0});

  return undef unless (defined($node));

  my $fs2 = {};
  my ($type);
  if (defined($type = $fs->theta($node)) && !ref($type)
      # re-bless class-names which inherit from us
      && UNIVERSAL::isa( $type, __PACKAGE__ )) {
    bless $fs2, $fs->theta($node);
  } else {
    bless $fs2, ref($fs);
  }
  %$fs2 = %$fs; # direct structure-sharing copy
  $fs2->root($node);
  return $fs2;
}

# $bool = $fs1->is_eq($fs2)
#   + underlying object-identity
sub is_eq {
  return
    (defined($_[0]) && defined($_[1]) &&
     $_[0]->{Qset} eq $_[1]->{Qset} &&
     $_[0]->{theta}->is_eq($_[1]->{theta}) &&
     $_[0]->{delta}->is_eq($_[1]->{delta}));
}

# $bool = $fs1->is_eql($fs2);
#   + underlying object-equivalence
sub is_eql {
  return 1 if (is_eq(@_));
  my ($fs1,$fs2) = @_;
  #return '' unless ($fs1->{Qset}->is_equal($fs2->{Qset}));
  my ($q1,$f,$val1,$val2);
  foreach $q (values(%{$fs1->{Qset}})) {
    return '' unless
      (exists($fs2->{$q})
       && ((!defined($fs1->theta($q)) && !defined($fs2->theta($q)))
	   || (defined($fs1->theta($q)) && defined($fs2->theta($q))
	       && $fs1->theta($q) eq $fs2->theta($q))));
    foreach $f ($fs1->features($q), $fs2->features($q)) {
      $val1 = $fs1->delta($q,$f);
      $val2 = $fs2->delta($q,$f);
      return '' unless
	((!defined($val1) && !defined($val2)) ||
	 (defined($val1) && defined($val2) &&
	  $val1 eq $val2));
    }
  }
  return 1;
}

# $fs2 = $fs1->pvalue(\@path); $fs1->pvalue(\@path,$val)
sub pvalue {
  return $_[0]->_get_pvalue($_[1]) if (scalar(@_) <= 2);
  return
    ($_[0]->can('gobble')
     ? $_[0]->_set_pvalue_gobble(@_[1,2])
     : $_[0]->_set_pvalue_nogobble(@_[1,2]));
}

sub _get_pvalue {
  my $fs = shift;
  my $path = shift;
  my $node = $fs->_get_nvalue($path);
  return undef unless (defined($node));
  return $fs->substructure($node);
  return $fs2;
}
sub _set_pvalue_gobble {
  my ($fs,$path,$val) = @_;

  # is the path empty ?
  unless ($path && @$path && defined($path->[0])) {
    $fs->gobble($val);
    return $fs;
  }

  # gobble it up
  my $val2node = {};
  $fs->gobble($val,
	      from => $fs->_get_or_create_nvalue([@$path[0..$#$path-1]]),
	      at => $path->[$#$path],
	      val2node => $val2node);
  return $fs->substructure($val2node->{$val});
}
# $fs->_set_pvalue_nogobble($path,$val)
sub _set_pvalue_nogobble {
  # we don't have gobble -- just set the type
  my ($node);
  $_[0]->_set_nvalue($_[1], ($_[0]->add_node($node = $_[0]->_gensym_node)));
  $_[0]->theta($node,$_[2]);
  return $_[0]->substructure($node);
}



# $roottype = $fs->theta();
# $type = $fs->theta($node);
# $fs->theta($node,$type)
# ALIASES: type
sub theta {
  return
    (defined($_[1]) # theta-keys must be defined!
     ? (exists($_[2])
	? $_[0]->{theta}->store($_[1],$_[2])
	: $_[0]->{theta}->fetch($_[1]))
     : $_[0]->{theta}->fetch($_[0]->root));
}
*type = \&theta;

# XXX $hashref = $fs->delta();
# XXX $hashref = $fs->delta($from_node)
# $to_node = $fs->delta($from_node,$ftr)
# $fs->delta($from_node,$ftr,$to_node)
# ALIASES: fvalue
*fvalue = \&delta;
sub delta {
  return
    (defined($_[1]) && defined($_[2])
     ? (exists($_[3])
	? (defined($_[3])
	   ? $_[0]->{delta}->store($_[1],$_[2],$_[3])
	   : $_[0]->{delta}->remove_transit($_[1],$_[2]))
	: $_[0]->{delta}->fetch($_[1],$_[2]))
     : undef);
}

# @feats = $fs->features()
# @feats = $fs->features($from_node);
sub features {
  return
    (defined($_[1])
     ? ($_[0]->{delta}->labels_from($_[1]))
     : ($_[0]->{delta}->alphabet));
}

###############################################################
# Information
###############################################################
# @paths = $fs->paths() : DANGEROUS
# @strings = $fs->pathstrings();
sub paths {
  return
    @{$_[0]->iterate_paths
	({
	  callback => \&_gather_paths_callback,
	  return => []
	 })};
}
sub _gather_paths_callback {
  push(@{$_[3]->{return}}, $_[1]);
  return undef;
}

# @hooks = $fs->hooks()
# @strings = $fs->hookstrings();
sub hooks {
  return
    @{$_[0]->iterate_hooks
	({
	  callback => \&_gather_paths_callback,
	  return => [],
	 })};
}

# $bool = $fs->has_path(\@ftrs);
sub has_path { return defined($_[0]->_get_nvalue($_[1])); }




# $bool = $fs->is_cyclic
sub is_cyclic {
  return
    $_[0]->iterate_paths({
			  callback => \&_is_cyclic_callback,
			  nodes => {}, # treated as set
			  return => '',
			 });
}
# callback($fs,$path,$node,$args)
sub _is_cyclic_callback {
  #return 1 if ($_[3]->{nodes}->member($_[2]));
  return 1 if (exists($_[3]->{nodes}{$_[2]}));
  $_[3]->{nodes}{$_[2]} = $_[2];
  return undef;
}

# $bool = $fs->is_reachable($to)
# $bool = $fs->is_reachable($to,$from)
sub is_reachable {
  my $fs = shift;
  my $to = shift;
  my $from = shift;
  $from = $fs->_master_root_node unless (defined($from));
  return '' unless
    (defined($from) && defined($to) && $fs->has_nodes($to, $from));
  return
    $fs->substructure($from)->iterate
      ({
	callback => \&_is_reachable_callback,
	find => $to,
	return => ''
       });
}
# callback($fs,$path,$node,$args)
sub _is_reachable_callback {
  return 1 if ($_[2] eq $_[3]->{find});
  return undef;
}

# @nodes = $fs->reachable_nodes()
# @nodes = $fs->reachable_nodes($node)
sub reachable_nodes {
  $_[1] = $_[0]->_master_root_node if (scalar(@_) == 1);
  return qw() unless ($_[0]->has_node($_[1]));
  return
    @{$_[0]->substructure($_[1])->iterate
	({
	  callback => \&_reachable_callback,
	  return => []
	 })};
}
# callback($fs,$path,$node,$args)
sub _reachable_callback {
  push(@{$_[3]->{return}}, $_[2]);
  return undef;
}

# @nodes = $fs->unreachable_nodes()
# @nodes = $fs->unreachable_nodes($node)
# ALIAS: dangling_nodes()
*dangling_nodes = \&unreachable_nodes;
sub unreachable_nodes {
  my %unodes = %{$_[0]->{Qset}};
  delete(@unodes{shift->reachable_nodes(@_)});
  return values(%unodes);
}

# $fs = $fs->remove_unreachable();
# $fs = $fs->remove_unreachable($node);
# ALIASES: remove_dangling(), trim()
*remove_dangling = \&remove_unreachable;
*trim = \&remove_unreachable;
*remove_unreachable_nodes = \&remove_unreachable;
sub remove_unreachable {
  return $_[0]->remove_nodes(shift->unreachable_nodes(@_));
}


###############################################################
# The necessary guts
###############################################################
# $fs->clear()
sub clear {
  my $fs = shift;
  #$fs->{Qset}->clear;
  %{$fs->{Qset}} = ();
  $fs->{q0} = undef;
  $fs->{theta}->clear;
  $fs->{delta}->clear;
  $fs->{qcounter} = 0;
  return $fs;
}

# $fs->clone()
sub clone { return $_[0]->new->assign($_[0]); }

# $dst_fs = $dst_fs->assign($src_fs)
sub assign {
  my ($dst,$src) = @_;
  my $tmp;
  $dst->hierarchy($src->hierarchy);
  #$dst->{Qset}->assign($src->{Qset});
  %{$dst->{Qset}} = %{$src->{Qset}};
  $dst->{q0} = $src->{q0};
  $dst->{theta}->assign($src->{theta});
  $dst->{delta}->assign($src->{delta});

  $dst->{qcounter} = $src->{qcounter};
  $dst->{pathsep} = $src->{pathsep};
  $dst->{pathre} = $src->{pathre};
  ${$dst->{master}} = ${$src->{master}};

  return $dst;
}

# $fs->alienate($fs2)
#   + renames nodes in $fs1 that exist in $fs2
sub alienate {
  my ($fs1,$fs2,$besafe) = @_;
  my (@common,%rename);

  # find common nodes
  # ARGH! continue here!
  @common = $fs1->{Qset}->intersection($fs2->{Qset})->members;
  return $fs1 unless (@common);

  # generate new node-names
  @rename{@common} = map { $fs1->_gensym_node($fs2) } @common;

  # rename the nodes
  $fs1->rename_nodes(%rename);

  return $fs1;
}
*make_node_disjoint = \&alienate;



###############################################################
# Iteration
#   + callback convention: ($fs, $path, $node, \%args)
###############################################################
sub iterate_paths {
  my $fs = shift;
  my $args = shift;
  return undef unless (defined($args) && defined($fs->root) && defined($args->{callback}));
  my $size = scalar($fs->nodes) * scalar($fs->features);
  my ($qe,$path,$node,$r);
  my @q = ( [[], $fs->root] ); # entries in queue @q are lists [ \@path, $node_at_path ]
  while ($qe = pop(@q)) {
    ($path,$node) = @$qe;
    $r = $args->{callback}($fs, $path, $node, $args);
    return $r if (defined($r));

    # check for cyclic structures
    --$size;
    if ($size < 0) {
      carp("Warning: iterate_paths() called for cyclic FStructure -- bailing out")
	if ($VERBOSE);
      last;
    }

    push(@q,
	 map {
	   [
	    [@$path, $_],          # path
	    $fs->delta($node,$_)   # path-value
	   ]
	 } $fs->features($node));
  }
  return $args->{return};
}


# iterate_hooks(\%args)
#  + depth-first hook-iteration
sub iterate_hooks {
  my $fs = shift;
  my $args = shift;
  return undef unless (defined($args) && defined($fs->root) && defined($args->{callback}));

  my $path = [];                # paths are lists of features
  my @pathq = ($path);          # @pathq is the queue of open paths
  my @nodeq = ($fs->root);      # @nodeq is the queue of open nodes

  # $ivals is a hashref of intermediate-path-values:
  #   { $path => Set::Hashed($node1,$node2,...), ... )
  # such that some proper prefix of $path has the path-value $node1, $node2, etc.
  # --> in other words, $ivals->{$path} is the set of nodes from which
  #     the node at $path is known to be reachable
  # --> this allows us to catch & identify the HOOKS.
  my %ivals = ( $path => Set::Hashed->new() );

  my ($node,$r,$newpath,$newnode);
  while ($path = pop(@pathq)) {
    $node = pop(@nodeq);
    $r = $args->{callback}($fs, $path, $node, $args);
    return $r if (defined($r));
    if (!$ivals{$path}->member($node)) {
      # still at a hook...
      foreach ($fs->features($node)) {
	$newnode = $fs->delta($node, $_);
	$newpath = [@$path, $_];
	# remember intermediate values
	$ivals{$newpath} = $ivals{$path}->clone->insert($node);
	push(@pathq, $newpath);
	push(@nodeq, $newnode);
      }
    }
  }
  return $args->{return};
}


# iterate_tracking(\%args)
#  + maintains the following keys in \%args
#    - $args->{open}
#       hashref of open nodes { node => 1 } if node has been opened
#        and not closed
#    - $args->{ivals}
#       hashref of intermediate pathvals
#        { $path => Set::Hashed, ... }
#       $args->{$ivals}{$path}->member($node) iff
#        some proper prefix of $path has $node as its path-value
#  + responds to the following keys in args:
#    - $args->{callback_open}
#       code-ref, called on path-open as
#        $args->{callback_open}($fs,$path,$node,\%args)
#    - $args->{callback_close}
#       code-ref, called on path-close as
#        $args->{callback_close}($fs,$path,$node,\%args)
#    - return rules as for $args->{callback} in
#      iterate(), etc.
#  + search-order and -paths otherwise as for
#    iterate_hooks().
sub iterate_tracking {
  my $fs = shift;
  my $args = shift;
  return undef unless (defined($args) && defined($fs->root) &&
		       (defined($args->{callback_open}) ||
			defined($args->{callback_close})));

  my $path = [];                # paths are lists of features
  my $node = $fs->root();       # current node under investigation

  # parallel stacks:
  my @pathstack = ($path);      # @pathstack is the stack of open paths
  my @nodestack = ($node);      # @nodestack is the stack of open nodes

  # $ivals is a hashref of intermediate-path-values:
  #   { $path => Set::Hashed($node1,$node2,...), ... )
  # such that some proper prefix of $path has the path-value $node1, $node2, etc.
  # --> in other words, $ivals->{$path} is the set of nodes from which
  #     the node at $path is known to be reachable
  # --> this allows us to catch & identify the HOOKS.
  my $ivals =
    defined($args->{ivals})
      ? $args->{ivals}
      : ($args->{ivals} = {});
  $ivals->{$path} = Set::Hashed->new();

  # $open is a Set::Hashed of open PATHS (listrefs)
  my $open =
    defined($args->{open})
      ? $args->{open}
      : ($args->{open} = {});

  my ($r,$newpath,$newnode);
  while ($path = pop(@pathstack)) {
    $node = pop(@nodestack);
    if ($open->{$path}) {
      # we've already opened this path: must be time to close it
      if (defined($args->{callback_close})) {
	$r = $args->{callback_close}($fs, $path, $node, $args);
	return $r if (defined($r));
      }
      delete($open->{$path});
      next;
    }
    # path is not already open: open it up
    if (defined($args->{callback_open})) {
      $r = $args->{callback_open}($fs, $path, $node, $args);
      return $r if (defined($r));
    }
    # remember we've opened it and stack it for close
    $open->{$path} = 1;
    push(@pathstack, $path);
    push(@nodestack, $node);

    # stack up its substructures (hooks only)
    if (!$ivals->{$path}->member($node)) {
      foreach ($fs->features($node)) {
	$newnode = $fs->delta($node, $_);
	$newpath = [@$path, $_];
	# remember intermediate values
	$ivals->{$newpath} = $ivals->{$path}->clone->insert($node);
	push(@pathstack, $newpath);
	push(@nodestack, $newnode);
      }
    }
  }
  return $args->{return};
}

# iterate_tracking_prefixes(\%args)
#  + like iterate_tracking(), additionally maintains
#    the following keys in \%args:
#    - $args->{closed}
#       hashref of closed nodes { $node => 1 } if $node has been
#        opened and closed
#    - $args->{prefixof}
#       hashref of paths' immediate prefixes:
#        { $path => $prefix}
#       $args->{prefixof}{$path} eq $prefix iff
#        $path is [ @$prefix, $ftr ], for some feature $ftr
#    - $args->{prefixes}
#       hashref of sets of paths' proper prefixes:
#        { $path => Set::Hashed, ... }
#       $args->{prefixes}{$path1}->member($path2)
#        iff $path2 is a proper prefix of $path1
#    - $args->{ivals}
#       hashref of intermediate pathvals
#        { $path => Set::Hashed, ... }
#       $args->{$ivals}{$path}->member($node) iff
#        some proper prefix of $path has $node as its path-value
sub iterate_tracking_prefixes {
  my $fs = shift;
  my $args = shift;
  return undef unless (defined($args) && defined($fs->root) &&
		       (defined($args->{callback_open}) ||
			defined($args->{callback_close})));

  my $path = [];                # paths are lists of features
  my $node = $fs->root();       # current node under investigation

  # parallel stacks:
  my @pathstack = ($path);      # @pathstack is the stack of open paths
  my @nodestack = ($node);      # @nodestack is the stack of open nodes

  # $ivals is a hashref of intermediate-path-values:
  #   { $path => Set::Hashed($node1,$node2,...), ... )
  # such that some proper prefix of $path has the path-value $node1, $node2, etc.
  # --> in other words, $ivals->{$path} is the set of nodes from which
  #     the node at $path is known to be reachable
  # --> this allows us to catch & identify the HOOKS.
  my $ivals =
    defined($args->{ivals})
      ? $args->{ivals}
      : ($args->{ivals} = {});
  $ivals->{$path} = Set::Hashed->new();

  # $open is a hashref from open PATHS (listrefs) to booles
  my $open =
    defined($args->{open})
      ? $args->{open}
      : ($args->{open} = {});

  # $closed is a hashref from closed PATHS (listrefs) to booles
  my $closed =
    defined($args->{closed})
      ? $args->{closed}
      : ($args->{closed} = {});

  # $prefixes is a hashref from paths (listrefs) to proper prefixes to booles
  my $prefixes =
    defined($args->{prefixes})
      ? $args->{prefixes}
      : ($args->{prefixes} = {});
  $prefixes->{$path} = Set::Hashed->new();

  # $prefixof is a hashref from paths (listrefs) to immediate prefixes (listrefs)
  my $prefixof =
    defined($args->{prefixof})
      ? $args->{prefixof}
      : ($args->{prefixof} = {});

  my ($r,$newpath,$newnode);
  while ($path = pop(@pathstack)) {
    $node = pop(@nodestack);
    if ($open->{$path}) {
      # we've already opened this path: must be time to close it
      if (defined($args->{callback_close})) {
	$r = $args->{callback_close}($fs, $path, $node, $args);
	return $r if (defined($r));
      }
      delete($open->{$path});
      $closed->{$path} = 1;
      next;
    }
    # path is not already open: open it up
    if (defined($args->{callback_open})) {
      $r = $args->{callback_open}($fs, $path, $node, $args);
      return $r if (defined($r));
    }
    # remember we've opened it and stack it for close
    $open->{$path} = 1;
    push(@pathstack, $path);
    push(@nodestack, $node);

    # stack up its substructures (hooks only)
    if (!$ivals->{$path}->member($node)) {
      foreach (reverse(sort($fs->features($node)))) {
	$newnode = $fs->delta($node, $_);
	$newpath = [@$path, $_];
	# remember prefixes
	$prefixes->{$newpath} = $prefixes->{$path}->clone->insert($path);
	$prefixof->{$newpath} = $path;
	# remember intermediate values
	$ivals->{$newpath} = $ivals->{$path}->clone->insert($node);
	push(@pathstack, $newpath);
	push(@nodestack, $newnode);
      }
    }
  }
  return $args->{return};
}


# $fs->iterate(\%args)
sub iterate {
  my $fs = shift;
  my $args = shift;
  return undef unless (defined($args) && defined($fs->root));

  my @q = ( [[], $fs->root] );  # entries in queue @q are lists [ \@path, $node ]

  # %checked is a hash ( $node => 1 ) to track which substructures
  # we've already iterated over
  my %checked = ();

  my ($qe,$path,$node,$r,$newnode);
  while ($qe = shift(@q)) {
    ($path,$node) = @$qe;
    $checked{$node} = 1;
    $r = $args->{callback}($fs, $path, $node, $args);
    return $r if (defined($r));
    push(@q,
	 map {
	   $newnode = $fs->delta($node, $_);
	   !defined($newnode) || $checked{$newnode}
	     ? qw()
	     : [
		[@$path, $_],   # path
		$newnode        # path-value
	       ]
	   } $fs->features($node));
  }
  return $args->{return};
}


###############################################################
# FS-Operations
###############################################################
sub subsumes {
  my ($fs1,$fs2) = @_;
  return undef if (!ref($fs1) || !ref($fs2)); # don't work as a class method.
  my $path = [];
  my $node1 = $fs1->root();
  my $node2 = $fs2->root();
  return 1 unless (defined($node1)); # unrooted fs's subsume *everything*
  return 0 unless (defined($node2)); # rooted fs's don't subsume unrooted ones

  # $checking is a hashref of node-pairs we're comparing:
  #   { $node1 => { $node2 => 1, ... } , ... }
  # s.t. $checking->{$node1}{$node2} if we've mapped $node1 to $node2
  my $checking = {};

  # $checked is like $checking, but contains already-finished node-pairs
  my $checked = {};

  my @nodeq = ($node1,$node2);
  my ($ftr,$new1,$new2);
  while (defined($node1 = shift(@nodeq))) {
    $node2 = shift(@nodeq);
    next if ($checked->{$node1}{$node2});
    $checking->{$node1}{$node2} = 1;

    # do type-subsumption check for this pair
    return '' unless ($fs1->_type_subsumes($fs1->theta($node1),
					   $fs2->theta($node2)));

    # if that's kosher, keep going...
    foreach $ftr ($fs1->features($node1)) {
      next unless (defined($new1 = $fs1->delta($node1,$ftr)));
      return '' unless (defined($new2 = $fs2->delta($node2,$ftr)));
      unless ($checking->{$new1}{$new2}) {
	push(@nodeq, $new1, $new2);
	$checking->{$new1}{$new2} = 1;
      }
    }
    $checked->{$node1}{$node2} = 1;
  }
  return 1;
}

sub extends { return subsumes($_[1],$_[0]); }

# $bool = $fs1->is_equal($fs2)
#   + logical equivalence
sub is_equal {
  return subsumes($_[0],$_[1]) && subsumes($_[1],$_[0]);
}


# $fs3_or_undef = $fs1->unify($fs2);
sub unify { return $_[0]->clone->_unify($_[1]); }

# $fs1_or_undef = $fs1->_unify($fs2);
sub _unify {
  my ($fs1,$fs2) = @_;

  # make fs's node-disjoint
  $fs1->alienate($fs2);

  # $classes is a hash of equivalence classes, indexed by node
  # equivalence classes are represented as Set::Hashed objects
  my $classes = $fs1->_merge_coreference($fs2);
  my $equiv = Set::Hashed->new(values(%$classes));

  # $class2node is a hash of node-names by equivalence-classes
  my %class2node =
    map { $_ => $fs1->_gensym_node($fs2) } $equiv->members;

  my ($class,$node,$q,$type);
  foreach $class ($equiv->members) {
    # iterate among classes, merrily constructing as we go...
    $node = $class2node{$class};
    $fs1->add_node($node);

    $type = undef;
    foreach $q ($class->members) {

      # update theta (!)
      if ($fs1->has_node($q)) {
	$type = $fs1->_type_join($type,$fs1->theta($q));
      } else {
	$type = $fs1->_type_join($type,$fs2->theta($q));
      }

      if (defined($type) && $type eq $TYPE_TOP) {
	# no join possible -- we can stop here
	$fs1->theta($node, $TYPE_TOP);
	$fs1->theta($fs1->root, $TYPE_TOP) if (defined($fs1->root));
	return undef;
      }

      # update delta function
      foreach $ftr ($fs1->features($q)) {
	$fs1->delta($node, $ftr, $class2node{$classes->{$fs1->delta($q,$ftr)}});
      }
      foreach $ftr ($fs2->features($q)) {
	$fs1->delta($node, $ftr, $class2node{$classes->{$fs2->delta($q,$ftr)}});
      }
    }
    # set the node-type
    $fs1->theta($node,$type);

    # remove old nodes
    $fs1->{Qset}->_difference($class);
    $fs1->{theta}->remove($class->members);
    $fs1->{delta}->remove_transits_from($class->members);
  }

  # set the new root node
  if (defined($fs1->{q0})) {
    $fs1->{q0} = $class2node{$classes->{$fs1->{q0}}};
    #$fs1->rename_nodes($class2node{$classes->{$fs1->root}}, $fs1->root); SLOW
    #$fs1->root($class2node{$classes->{$fs1->root}}); : NOPE
  } elsif (defined($fs2->{q0})) {
    $fs1->{q0} = $class2node{$classes->{$fs2->{q0}}};
    #$fs1->rename_nodes($class2node{$classes->{$fs2->root}}, $fs2->root);
  }

  # (re)set the master root
  if (defined(${$fs1->{master}})) {
    ${$fs1->{master}} = $class2node{$classes->{${$fs1->{master}}}};
  } elsif (defined(${$fs2->{master}})) {
    ${$fs1->{master}} = $class2node{$classes->{${$fs2->{master}}}};
  }

  return $fs1;
}
*extend = \&_unify;


sub _merge_coreference {
  my ($fs1,$fs2) = @_;

  # $classes is basically a hash of equivalence classes, indexed by fs-node
  # equivalence classes are represented as Set::Hashed objects
  # -- the trick: guarantee set-equivalence by using an FStructure
  #               to represent the equivalence classes, since these
  #               can already handle structure-sharing
  my $classes = {};
  %$classes =
    map {
      $_ => Set::Hashed->new($_)
    } $fs1->{Qset}->union($fs2->{Qset})->members;

  return $classes unless (defined($fs1->root) &&
			  defined($fs2->root));

  # $pairs is a hash of identified node-pairs
  # { $node1 => { $node2 => 1, ...}, ... }
  my $pairs = { $fs1->root => { $fs2->root => 1 } };

  my ($q1,$q2,$s1,$s2,$ftr,$new);
  while (%$pairs) {

    # pick a (random) pair of nodes
    $q1 = (keys(%$pairs))[0];
    $q2 = (keys(%{$pairs->{$q1}}))[0];

    # remove them from the pair-queue
    delete($pairs->{$q1}{$q2});
    delete($pairs->{$q1}) unless (%{$pairs->{$q1}});

    # get equivalence-classes for nodes
    $s1 = $classes->{$q1};
    $s2 = $classes->{$q2};

    # check object-identity
    next if ($s1 eq $s2);

    # check set-identity & ensure object-identity
    if ($s1->is_equal($s2)) {
      foreach ($s2->_difference($s1)->members) {
	$classes->{$_} = $s1;
      }
      next;
    }

    # we must needs recurse
    foreach ($s2->difference($s1)->members) {
      $classes->{$_} = $s1;
    }
    $s1->_union($s2);

    foreach $ftr ($fs1->features($q1)) {
      next unless (defined($new = $fs2->delta($q2,$ftr)));
      $pairs->{$fs1->delta($q1,$ftr)}{$new} = 1;
    }
  }
  return $classes;
}


###############################################################
# Type-subsumption stuff
###############################################################

# $bool = $fs->_type_subsumes($t1,$t2)
sub _type_subsumes ($$$) {
  return
    (shift->hierarchy or 'Math::PartialOrder')->subsumes($_[0],$_[1]);
}

# $bool = $fs->_type_extends($t1,$t2)
sub _type_extends ($$$) {
  return
    (shift->hierarchy or 'Math::PartialOrder')->extends($_[1],$_[0]);
}

# $type = $fs->_type_join(@types);
sub _type_join ($@) {
  return
    (shift->hierarchy or 'Math::PartialOrder')->njoin(@_);
}

# $type = $fs->_type_meet(@types);
sub _type_meet ($@) {
  return
    (shift->hierarchy or 'Math::PartialOrder')->nmeet(@_);
}


1;
__END__

###############################################################
# Docs
###############################################################
# -*- Mode: Perl -*-
=pod

=head1 NAME

Data::FStructure::Base -
Feature structures for perl -- abstract base class.

=head1 SYNOPSIS

  use Data::FStructure::Base;

  #
  # Construction and Initialization
  #
  $fs = Data::FStructure::Base->new(\%args); # return a new feature-structure
  $fs2 = $fs1->new(\%args);                 # new empty FS, shared hierarchy

  #
  # Feature-Structure Operations
  #
  $fs2 = $fs->substructure($node);   # get substructure at $node
  $bool = $fs1->is_eq($fs2);         # test identity of underlying objects
  $bool = $fs1->is_eql($fs2);        # test equivalence of underlying objects

  $val = $fs->pvalue(\@path);        # get path-value for \@path
  $fs->pvalue(\@path, $val);         # set path-value for \@path

  $bool = $fs1->subsumes($fs2);      # test subsumption
  $bool = $fs1->extends($fs2);       # test reverse-subsumption
  $bool = $fs1->is_equal($fs2);      # test alphabetic variance

  $fs3 = $fs1->unify($fs2);          # unification
  $fs1->_unify($fs2);                # destructive unification

  $fs3 = $fs1->generalize($fs2);    # (NOT YET IMPLEMENTED)
  $fs->_generalize($fs2);           # (NOT YET IMPLEMENTED)

  #
  # Feature-Structure Manipulation & Information
  #
  $h = $fs->hierarchy();             # get type-hierarchy
  $fs->hierarchy($h);                # set type-hierarchy

  $node = $fs->root();               # get root-node
  $fs->root($node);                  # set root-node

  $bool = $fs->has_nodes(@nodes);    # are all @nodes present?
  @ids = $fs->add_nodes(@nodes);     # add nodes (new @ids for 'undef' values
  $fs->remove_nodes(@nodes);         # remove all @nodes
  $fs->rename_nodes(%old2new);       # rename nodes

  @nodes = $fs->nodes();              # list of defined nodes
  @nodes = $fs->reachable_nodes();    # list of all reachable nodes
  @nodes = $fs->unreachable_nodes() ; # list of all unreachable nodes
  $fs->remove_unreachable();          # removes unreachable nodes
  $bool = $fs->is_reachable($n1,$n2); # test node-reachability

  $type = $fs->type($node);           # get node-type
  $fs->type($node,$type);             # set node-type

  $node2 = $fs->fvalue($qfrom,$ftr);  # get feature-value for $ftr from $node1
  $fs->fvalue($qfrom,$ftr,$qto);      # set feature-value for $ftr from $node1

  @feats = $fs->features();           # get all features with defined values
  @feats = $fs->features($node);      # get features defined for $node

  @paths = $fs->paths();              # get feature-paths
  @hooks = $fs->hooks();              # get non-redundant paths
  $bool = $fs->has_path(\@path);      # is \@path defined in $fs?
  $bool = $fs->is_cyclic();           # test for self-referential paths

  $fs->clear();                       # clear all but hierarchy information
  $fs2->assign($fs1);                 # make $fs2 structure-identical to $fs1
  $fs2 = $fs1->clone();               # create a new distinct copy

  $fs->alienate($fs2);                # make $fs1 and $fs2 node-disjoint


  #
  # Iteration Utilities
  #
  $fs->iterate(\%args);                   # iterate over all nodes
  $fs->iterate_paths(\%args);             # iterate over all paths
  $fs->iterate_hooks(\%args);             # iterate over all hooks
  $fs->iterate_tracking(\%args);          # extended hook-iteration
  $fs->iterate_tracking_prefixes(\%args); # further-extended hook-iteration

  #
  # Low-Level Utilities
  #
  $bool = $fs->_type_subsumes($t1,$t2);   # type-subsumption check
  $bool = $fs->_type_extends($t1,$t2);    # reverse type-subsumption check
  $t3 = $fs->_type_join(@types);          # type-join operation
  $t3 = $fs->_type_meet(@types);          # type-meet operation
  $hash = $fs1->_merge_coreference($fs2); # node-identification algorithm

  #
  # Internal API
  #
  $Q = $fs->Qset();                       # set of nodes
  $q0 = $fs->q0();                        # root node handle
  $theta = $fs->_theta();                 # type map
  $delta = $fs->_delta();                 # transition function
  $noderef = $fs->_master();              # master root (reference)
  $node = $fs->_master_q0();              # master root (value)

=head1 REQUIREMENTS

Carp, Math::PartialOrder, Set::Hashed.

=cut

#--------------------------------------------------------------
# Description
#--------------------------------------------------------------
=pod

=head1 DESCRIPTION

The C<Data::FStructure::Base> module provides
an abstract base-class for FStructure objects which is
based on perl hashes.  All of the
elementary FStructure operations are defined in this package,
from which all other FStructure objects inherit.

It requires its descendants only to define their
C<theta> and C<delta> keys as (descendants of)
C<Data::FStructure::theta> and C<Data::FStructure::delta> objects, respectively.

=cut

#--------------------------------------------------------------
# Variables
#--------------------------------------------------------------
#=pod
#
#=head1 PACKAGE VARIABLES
#
#None.
#
#=cut

#--------------------------------------------------------------
# Methods
#--------------------------------------------------------------
=pod

=head1 METHODS

=cut


#--------------------------------------------------------------
# Construction
#--------------------------------------------------------------
=pod

=head2 Construction and Initialization

=over 4

=item * C<new(\%args)>

Returns a new Data::FStructure object.

\%args is a hashref for which C<Data::FStructure::Base>
recognizes the following keys:

=over 4

=item * C<< hierarchy =E<gt> $h >>

Sets the initial hierarchy for this FStructure to $h,
which should be a descendant of Math::PartialOrder.
If new() is called as a class method,
C<< Data::FStructure-E<gt>new() >>
then $h defaults to C<undef> (no hierarchy),
so that the returned FStructure object
will behave as an untyped feature structure.

If new() is called as an instance method,
C<< $fs-E<gt>new() >>, then $h defaults to the
hierarchy associated with $fs, if any.

=item * C<< root =E<gt> $node >>

Sets the initial root for the FStructure to C<$node>.
By default, FStructures created by the C<new()> method
have no nodes at all.


=item * C<< qcounter =E<gt> $i >>

If C<qcounter> is specified, it will be used to name new
nodes (names generated by C<undef> values passed to the
C<add_nodes()> method).
If C<new()> was called as an
object-method, $i defaults to the current counter-value
of the prototype.  Otherwise, it defaults to 0.

=item * C<< theta =E<gt> $th >>

Sets the type-mapping object for the new FStructure to
C<$th>, which should be a C<Data::FStructure::theta> object.
Default: none.

=item * C<< delta =E<gt> $dt >>

Sets the type-mapping object for the new FStructure to
C<$dt>, which should be a C<Data::FStructure::delta> object.
Default: none.

=back

Subclasses may inherit the C<new()> method.  Subclass-specfic
initialization will be performed if the subclass defines or
inherits a method C<initialize()>, which will be called as:

  $fs->initialize(\%args);

=back

=cut


#--------------------------------------------------------------
# Feature-Structure Opertations
#--------------------------------------------------------------
=pod

=head2 Feature-Structure Operations

=over 4

=item * C<< $fs-E<gt>substructure($node) >>

If $fs has no node named $node, returns C<undef>.  Otherwise,
returns an FStructure object B<exactly> like $fs, but rooted at
$node.
This means that destructive operations (including C<assign()>)
on the value returned by this method will affect $fs itself.
If this is not desired, the C<clone()> method may be used:

  $safe = $fs->substructure($node)->clone();

If C<$node> is assigned a non-reference type which is the
name of a class inheriting from C<Data::FStructure::Base> via
perl's C<@ISA> mechanism, the returned substructure will be
blessed into that class.  Otherwise, it will be blessed
into whichever class C<$fs> belongs to.

=item * C<< $fs1-E<gt>is_eq($fs2) >>

Returns a true value if $fs1 and $fs2 share object-identical
(C<eq>) sets of nodes, type maps, and feature-transition maps,
false otherwise.

Note that this is B<not> the same as C<is_equal()>, which tests
for logical feature-structure equivalence.  Rather, this method
tests for object-equality of the underlying representations.

This method will, for example, return a true value if
$fs2 was created by a call to the C<substructure()> method
of $fs1, or vice versa, but may return false for
logically equivalent FStructures.  Similarly,
FStructures created with the C<substructure($node)>
will in the great majority of cases
(C<< $node ne $fs-E<gt>root >>) not be logically
equivalent, but will refer to the same underlying
representations.

This method is used (if the L<Data::FStructure::perldata> module is loaded)
for efficient substructure-assignment.

See also: C<substructure()>, C<is_equal>, L<Data::FStructure::perldata>.



=item * C<< $fs1-E<gt>is_eql($fs2) >>

Like C<is_eq>, but also returns a true value if
the underlying representations
(nodes, type-maps, and feature-transition functions)
are equivalent but not object-identical.

This method will return true for FStructures created
with any combination of the C<clone()> and C<substructure()>
methods.



=item * C<< $fs-E<gt>pvalue(\@path) >>, C<< $fs-E<gt>pvalue(\@path,$val) >>

With no arguments, get the path-value for the feature-path \@path in $fs.
The value returned is an instance of Data::FStructure
B<exactly> like $fs except for the value of the root node.  Same
caveats as for the C<substructure()> method.

With an argument, sets the path-value for the feature-path \@path
in $fs.  If the C<Data::FStructure::perldata> module is loaded, then
$val is C<gobbled()> into the FStructure at the specified path.
Otherwise, a new node is created in the FStructure, and is
assigned C<$val> as its type.

See also: The description of the C<gobble()> method in L<Data::FStructure::perldata>.



=item * C<< $fs1-E<gt>subsumes($fs2) >>

Returns a true value
if $fs1 subsumes (read: "is at least as general as") $fs2,
false otherwise.  A feature structure $fs1 subsumes a feature
structure $fs2 if the following hold:

=over 4

=item 1.

Every path with a defined value in $fs1 is defined in $fs2, and

=item 2.

For each path $path defined in $fs1, the type assigned by $fs1
to the node reached by following $path subsumes the type assigned by
$fs2 to the node reached by following the same $path.

=back

Path-identification is perfomed iteratively, and should handle
cyclic FStructures correctly.  Type-subsumption checks are performed
by a call to C<< $fs1-E<gt>_type_subsumes() >>.



=item * C<< $fs1-E<gt>extends($fs2) >>

Returns a true value
if $fs1 extends (read: "is at least as specific as") $fs2,
false otherwise.
Really just an alias for:

  $fs2->subsumes($fs1);



=item * C<< $fs1-E<gt>is_equal($fs2) >>

Returns a true value if $fs1 and $fs2 are alphabetic variants.
(read: "logically equivalent"), false otherwise.

See also: C<is_eq()>, C<is_eql>.



=item * C<< $fs1-E<gt>unify($fs2) >>

Returns a new feature structure representing the unification of
$fs1 and $fs2, if it can be constructed, otherwise returns C<undef>.
The unification of $fs1 and $fs2 contains all paths defined in
either $fs1 or $fs2.  The unification operation can be understood
as logically conjoining the information contained in $fs1 and
$fs2.

Each node of the returned feature structure is identified
by an equivalence class of nodes from $fs1 and $fs2, in order
to properly handle instances of structure-sharing.

Each path $path is assigned as its type the
least upper bound of the type it is assigned in $fs1 and $fs2.
Least upper bounds are computed by the
C<Math::PartialOrder> method C<njoin()>.
If no least upper bound can be
found for a given pair of paths, then the
unification fails, and C<undef> is returned.

Non-deterministic hierarchies currently cause
the C<njoin> method to produce a warning message and
select a (possibly unpredictable) default.
See L<Math::PartialOrder> for details.



=item * C<< $fs1-E<gt>_unify($fs2) >>, C<< $fs1-E<gt>extend($fs2) >>

Attempts to unify $fs1 with $fs2 by destructively altering $fs1
into its unification with $fs2, if it exists.  Returns $fs1
on success and C<undef> on failure.

If the unification fails, both the root-node of $fs1 and
the node causing the failed join will be assigned the
type C<$Math::PartialOrder::TYPE_TOP>.


=item * C<< $fs1-E<gt>generalize($fs2) >>

Not yet implemented.

=item * C<< $fs-E<gt>_generalize($fs2) >>

Not yet implemented.

=back

=cut


#'
#--------------------------------------------------------------
# Feature-Structure Manipulation & Information
#--------------------------------------------------------------
=pod

=head2 Feature-Structure Manipulation and Information

=over 4

=item * C<hierarchy()>, C<hierarchy($h)>

Get/set the type-hierarchy associated with the FStructure instance;
$h should be a descendant of Math::PartialOrder.


=item * C<root()>, C<root($node)>

Aliases: C<q0()>

Get/set the root-node of the FStructure.


=item * C<has_node($node)>, C<has_nodes(@nodes)>

C<has_nodes()> returns a true value if the FStructure
contains all of the nodes @nodes, false otherwise.
C<has_node()> tests only its initial argument.

=item * C<add_node()>, C<add_node($node)>

Adds a new node to the FStructure and returns its
name.  If no C<$node> argument is given, or if
C<$node> is not defined, a new node-name is generated.
Node-names default to the string representations of
the nodes added.  You may add any defined perl scalar
as a node, but if you want references, it is safer to
use the type-interface.

It is safe to call this method more than once for
defined C<$node> values.

=item * C<add_nodes(@nodes)>

Like C<add_node()>, but adds all of the nodes in
C<@nodes>, and returns their names as a list.


=item * C<remove_nodes(@nodes)>, C<remove_node(@nodes)>, C<remove(@nodes)>

Removes all nodes in the list @nodes, along with any types assinged
to them and any feature-transitions defined to them or from them.
Returns the FStructure.


=item * C<< rename_nodes($old =E<gt> $new, ...) >>

Aliases: C<rename_node()>, C<replace()>

Renames the C<$old> nodes to the corresponding C<$new> values.
Destructively alters node-set, type-map and feature-transitions,
as well as the root-node, if it is specified by some C<$old>
value.  Returns the FStructure.



=item * C<nodes()>

Aliases: C<Q()>.

Returns a list of all nodes defined in the FStructure.


=item * C<reachable_nodes()>, C<reachable_nodes($node)>

Returns a list of all nodes reachable from $node
by some (possibly empty) feature-path.  $node defaults
to the root of the FStructure.



=item * C<unreachable_nodes()>, C<unreachable_nodes($node)>

Alias: C<dangling_nodes()>.

Returns a list of all nodes not reachable from $node
by any (possibly empty) feature-path.  $node defaults
to the root of the largest superstructure of the FStructure.

See Also: L<DANGLING NODES>.


=item * C<remove_unreachable()>, C<remove_unreachable($node)>

Aliases: C<remove_dangling()>, C<trim()>.

Removes any and all references to those nodes of the FStructure
which cannot be reached from $node, which defaults to the root
node of the largest superstructure of the FStructure.
Returns the FStructure.

See Also: L<DANGLING NODES>.


=item * C<is_reachable($to)>, C<is_reachable($to,$from)>

Returns a true vale if the node $to is reachable from the node $from, false
otherwise.  A node $to is considered "reachable" from another
node $from if the extended feature-transition function leads from
$from to $to by some (possibly empty) feature-path.  $from defaults
to the root of the FStructure.



=item * C<type()>, C<type($node)>, C<type($node,$type)>

Alias: C<theta()>

With no arguments, gets the type assigned to the root node
of the FStructure, if any.  With a single argument, gets the
type assigned to $node.  With 2 arguments, assigns the type
$type to $node.

See Also: C<Data::FStructure::theta>.


=item * C<fvalue($node,$ftr)>, C<fvalue($node1,$ftr,$node2)>

Alias: C<delta>

With 2 arguments, gets the transition-value of the feature $ftr for
the node $node.  With 3 arguments, sets the transition-value for the feature
$ftr at $node1 to $node2.

See Also: C<Data::FStructure::delta>.


=item * C<nvalue(\@path)>, C<nvalue(\@path,$node)>

Alias: C<delta_star>

With 1 argument, get the extended transition value for the
feature-path \@path in the FStructure.

With 2 arguments, set the the extended transition value for
the feature-path \@path to $node.  $node is added to the
FStructure if it is not already defined, and intermediate
path-values are implicitly created.  Returns $node.




=item * C<features()>, C<features($node)>

With no arguments, returns a list of all features
defined in the feature-structure.  With 1 argument,
returns a list of the features defined for $node.

See Also: L<Data::FStructure::delta>.


=item * C<paths()>

Returns a list of all paths defined in the FStructure.
If the feature structure contains cyclic paths, this method
will produce a warning message and terminate.

See Also: L<iterate_paths(\%args)>.

=item * C<hooks()>

Returns a list of all non-redundant paths defined in
the FStructure.

See Also: L<iterate_hooks(\%args)>.


=item * C<has_path(\@path)>

Returns a true value if the feature-path \@path is defined
in the FStructure, false otherwise.


=item * C<is_cyclic()>

Returns a true value if the FStructure contains at least
one cyclic path, false otherwise.  A "cyclic path" is a
path whose node-value is identical (C<eq>) to that of
at least one of its proper prefixes.



=item * C<clear()>

Clears all information in the FStructure except for
the hierarchy.  To un-associate a hierarchy with an
FStructure, do

  $fs->hierarchy(undef);


=item * C<< $fs2-E<gt>assign($fs1) >>

Makes $fs2 a distinct but structurally identical copy of $fs1.
The hierarchy associated with $fs1, as well as type-values
are "copied" into $fs2 via native perl assignment -- references
(i.e. hierarchies) therefore will be shared.
Returns $fs2.


=item * C<< $fs-E<gt>clone() >>

Returns a new FStructure distinct from but structurally identical
to $fs.
The hierarchy associated with $fs, as well as type-values
are "copied" into the clone via native perl assignment -- references
(for instance hierarchies, as well as type- and node-values, if
they are references) will therefore be shared.


=item * C<< $fs-E<gt>alienate($fs2) >>

Alias: C<make_node_disjoint>

Renames any nodes in $fs which also exist in $fs2, assigning
them names that are not associated with any node in $fs2.
Returns $fs.  This method is used by the C<unify()> method
to ensure alphabetic variance.  It will currently drop
even nodes identified by references which are shared between
FStructures.  If you wish to store references in an
FStructure, it is safer to use the C<type()> interface
to do so.


=back

=cut

#'
#--------------------------------------------------------------
# Iteration Utilities
#--------------------------------------------------------------
=pod

=head2 Iteration Utilities

=over 4

=item * C<iterate(\%args)>

Iterates over the calling FStructure object C<$fs>, calling
C<&$args{callback}($fs, $path, $node, \%args)> for each path
C<$path> in C<$fs>, where C<$node> is the (extended) transition
value for C<$path> in C<$fs>.

If for any call C<&$args{callback}(...)> returns
a value other than C<undef>, the iteration is broken off and this
value is returned immediately.
Otherwise, returns the value of $args{return} at the end of
the iteration.

If the feature structure contains cyclic paths, this method
will produce a warning message and terminate.


=item * C<iterate_paths(\%args)>

Like C<iterate()>, but attempts to call
C<&{$args{callback}}> for every path defined in
the FStructure.  If this method is called for a cyclic
FStructure object, a warning will be issued (with C<carp()>)
and the method will return after having visited each
node at least once.  For possibly cyclic FStructure
objects, it is better to use an iteration method
such as C<iterate_hooks()>.


=item * C<iterate_hooks(\%args)>

Like C<iterate_paths()>, but $args{callback} is only called
for non-redundant paths.  A path is "redundant" whose node-value is
identical to the node-value of at least one of its proper prefixes.


=item * C<< $fs-E<gt>iterate_tracking(\%args) >>

Like C<iterate_hooks()>, but additionally maintains the following
keys in \%args:

=over 4

=item * C<< open =E<gt> $hash >>

$hash is a hash of open nodes to booleans: $hash{$node} == 1 if
the node has been opened and not yet closed.

=item * C<< ivals =E<gt> $hash >>

$hash is a hashref of intermediate path-values:

  { $path => Set::Hashed, ... }

C<< $hash{$path}-E<gt>member($node) >> will be true if and only if some proper
prefix of $path leads to $node in $fs.

=back

Also, C<iterate_tracking()> responds to the following keys in \%args:

=over 4

=item * C<< callback_open =E<gt> $code >>

  $code should be a coderef which is called on path-open with the
  same calling conventions as $args{callback}.

=item * C<< callback_close =E<gt> $code >>

  $code should be a coderef which is called on path-close
  with the same calling conventions as $args{callback}.

=back


=item * C<< $fs-E<gt>iterate_tracking_prefixes(\%args) >>

Like C<iterate_tracking>, but additionally maintains the
following keys in \%args:

=over 4

=item * C<< closed =E<gt> $hash >>

A hashref of nodes which have been closed to
booleans: C<< $hash-E<gt>{$node} >> is true if $node has been
opened and closed.

=item * C<< prefixof =E<gt> $hash >>

A hashref of paths' immediate prefixes of the form

  { $path => $prefix }

C<< $hash-E<gt>{$path} eq $prefix >> iff $prefix is the longest proper
prefix of $path.

=item * C<< prefixes =E<gt> $hash >>

Hashref of sets of paths' proper prefixes:

  { $path => Set::Hashed, ... }

C<< $hash{$path}-E<gt>member($prefix) >> iff $prefix is a proper
prefix of $path.

=back

=back

=cut

#'
#--------------------------------------------------------------
# Low-Level Utilities
#--------------------------------------------------------------
=pod

=head2 Low-Level Utilities

=over 4

=item * C<< $fs-E<gt>_type_subsumes($t1,$t2) >>

This method is really just a wrapper for
C<Math::PartialOrder::subsumes($t1,$t2)>, which
will be called as an instance method of
the hierarchy-instance associated with C<$fs>,
if it is defined, and otherwise as a class method.

See Also: C<Math::PartialOrder>.



=item * C<< $fs-E<gt>_type_join(@types) >>

This method is really just a wrapper for
C<Math::PartialOrder::njoin(@types)>, which
will be called as an instance method of
the hierarchy-instance associated with C<$fs>,
if it is defined, and otherwise as a class method.

See Also: C<Math::PartialOrder>.



=item * C<< $fs-E<gt>_type_meet(@types) >>

Like C<_type_join(@types)>, but wraps
C<Math::PartialOrder::nmeet(@types)>.

See Also: C<Math::PartialOrder>.



=item * C<< $fs1-E<gt>_merge_coreference($fs2) >>

Assumes that $fs1 and $fs2 share no node-names.
Returns a hash of Set::Hashed objects keyed by node-name,
in which the values are equivalence-classes for the nodes
in $fs1 and $fs2.  Nodes are considered equivalent which are
reachable via identical paths.  Set-identical equivalence
classes are object-identical (C<eq>) upon completion of this
method.

This method is an adaptation of the algorithm presented
by Ait-Kaci (1984).

See also: C<alienate()>.

=back

=cut


#--------------------------------------------------------------
# API
#--------------------------------------------------------------
=pod

=head2 API

The methods described describe the API of the underlying representation
of FStructure objects assumed by the abstract class
C<Data::FStructure::Base>.  They should never need to be called by the user.

=over 4

=item * C<Qset()>

Get/set the C<Set::Hashed> object representing all currently defined
nodes of the FStructure.

=item * C<q0()>, C<q0($node)>

Get/set the current root node of the FStructure.
Really, the C<root()> method is just an alias for C<q0()>.

=item * C<_theta()>, C<_theta($th)>

I<(Abstract)>

Get/set the type-mapping object for the FStructure, which
should be a C<Data::FStructure::theta> object.  This
method is not defined in C<Data::FStructure::Base>, but
should be defined by concrete subclasses.

See Also: L<Data::FStructure::theta>.


=item * C<_delta()>, C<_delta($dt)>

I<(Abstract)>

Get/set the transition-mapping object for the FStructure, which
should be a C<Data::FStructure::delta> object.  This
method is not defined in C<Data::FStructure::Base>, but
should be defined by concrete subclasses.

See Also: L<Data::FStructure::delta>.


=item * C<_master()>

Returns a reference to the root node of the largest
superstructure of the FStructure.  This reference
is maintained by C<root()> and C<unify()> operations,
and shared by FStructure objects produced by the
C<substructure()> method.  It is used by
the C<remove_unreachable_nodes()> method.


=item * C<_master_q0()>, C<_master_q0($node)>

Get/set the name of the master root node.
This method just sets the value of the same scalar-reference
returned by C<_master()>.

=back

=cut

#'
#--------------------------------------------------------------
# Danging nodes
#--------------------------------------------------------------
=pod

=head1 DANGLING NODES

Due to the recursive nature of feature structures and the
(comparatively fast) implementation of substructures as
new "handles" into the embedding FStructures, it is possible
that destructive FStructure operations (path- and node-value
assignment in particular) will create "dangling" nodes in the
FStructure objects upon which they operate.

This means that perl's
internal reference count to such dangling nodes will never
reach zero (and that they will continue to occupy memory)
as long as they remain in the FStructure, assuming of course
that a reference to the FStructure still exists.

The method C<remove_unreachable_nodes()>
provides the only automated mechanism currently defined for
removing such dangling nodes.  Since
C<remove_unreachable_nodes()> must iterate over every node
in the embedding FStructure, repeated calls may lead to a
significant decrease in efficiency for most applications.

For this reason, C<remove_unreachable_nodes()>
is never called by any of the other API methods
described above, and users wishing to save memory
space must manually call it as needed,
or use a garbage-collecting descendant such as
C<Data::FStructure::Safe>, or one of its descendants
such as C<Data::FStructure>.

Future versions of this module may implement a
perl-esque reference-counting strategy to keep track
of and do away with dangling nodes as they cease
to be reachable.

=cut

#'
#--------------------------------------------------------------
# Footer
#--------------------------------------------------------------
=pod

=head1 BUGS

It should work directly with perl data-structures, i.e. references
instead of node-names.

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).
Data::FStructure(3pm).
Data::FStructure::Safe(3pm).
Data::FStructure::Std(3pm).
Data::FStructure::Array(3pm).
Data::FStructure::List(3pm).
Data::FStructure::perldata(3pm).
Data::FStructure::theta(3pm).
Data::FStructure::delta(3pm).
Data::FStructure::arydelta(3pm).
Data::FStructure::pretty(3pm).
Math::PartialOrder(3pm).
Set::Hashed(3pm).
