# -*- Mode: Perl -*-

#
# Copyright (c) 2001, Bryan Jurish.  All rights reserved.
#
# This package is free software.  You may redistribute it
# and/or modify it under the same terms as Perl itself.
#

###############################################################
#
#      File: Data::FStructure::Safe.pm
#    Author: Bryan Jurish <jurish@ling.uni-potsdam.de>
#
# Description: Garbage-collecting Feature-structures for perl
#
###############################################################

package Data::FStructure::Safe;
# System modules
use Carp;
require Exporter;
# 3rd party exstensions
# user extension modules
use Data::FStructure::Base;
@ISA       = qw(Data::FStructure::Base);
@EXPORT   = qw();
@EXPORT_OK = (
	      qw($GC_INTERVAL $GC_VERBOSE)
	     );
%EXPORT_TAGS =
  (
   gcvars => [qw($GC_INTERVAL $GC_VERBOSE)]
  );

###############################################################
# Package-global variables
###############################################################
our $VERSION = 0.02;
our $GC_INTERVAL = 1024;
our $GC_VERBOSE = 0;


###############################################################
# Constructor
#   + Data::FStructure::Safe->new( {hierarchy=>$h} )
#   + Object-structure:
#     { --- inherited ---
#       ... a buncha stuff...
#       --- new ---
#      gc_interval   => Scalar num/destructive ops until gc
#      gc_counter    => REF to Scalar num/destructive ops performed
#      gc_ignore     => Scalar boolean flag (for unification)
#     ]
#   + New Args:
#     {
#       gc_interval => $int,
#       gc_counter => $int,
#       gc_dont => $bool
#     }
###############################################################
sub initialize {
  my $self = shift;
  my $args = shift;
  my $ctr = $args->{gc_counter} || 0;
  $self->{gc_counter} = \$ctr unless (defined($self->{gc_counter}));
  $self->{gc_interval} = ($args->{gc_interval}
			  || $self->_gc_interval
			  || $GC_INTERVAL);
  $self->{gc_dont} = ($args->{gc_dont} || 0);
  return $self;
}


###############################################################
# Accessors/manipulators
###############################################################
sub _gc_interval {
  return $_[0]->{gc_interval} if (!exists($_[1]));
  $_[0]->{gc_interval} = $_[1];
  $_[0]->_check_gc() unless ($_[0]->{gc_dont});
  return $_[1];
}

sub _gc_counter {
  return ${$_[0]->{gc_counter}} if (!exists($_[1]));
  return ${$_[0]->{gc_counter}} = $_[1];
}
sub _gc_counter_ref {
  return $_[0]->{gc_counter} if (!exists($_[1]));
  return $_[0]->{gc_counter} = $_[1];
}


###############################################################
# Low-level utilities
###############################################################
# _check_gc(): do garbage-collection check
sub _check_gc {
  unless (${$_[0]->{gc_counter}} >= $_[0]->{gc_interval}) {
    return ++${$_[0]->{gc_counter}};
  }
  print STDERR ("Garbage collection in progress for $_[0] ...") if ($GC_VERBOSE);
  $_[0]->remove_unreachable();
  print STDERR (" completed.\n") if ($GC_VERBOSE);
}

###############################################################
# Basic feature-structure manipulation
###############################################################


# $to_node = $fs->delta($from_node,$ftr)
# $fs->delta($from_node,$ftr,$to_node)
# ALIASES: fvalue
*fvalue = \&delta;
sub delta {
  return $_[0]->SUPER::delta(@_[1..$#_]) if (!exists($_[3]) || $_[0]->{gc_dont});
  my $rv = $_[0]->SUPER::delta(@_[1..3]);
  $_[0]->_check_gc;
  return $rv;
}


###############################################################
# Information
###############################################################


###############################################################
# The necessary guts
###############################################################
# $fs->clear()
sub clear {
  my $fs = shift;
  $fs->SUPER::clear();
  $fs->_gc_counter(0);
  return $fs;
}

# $fs = $fs->assign($fs2)
sub assign {
  my ($dst,$src) = @_;
  $dst->{gc_interval} = $src->{gc_interval};
  ${$dst->{gc_counter}} = ${$src->{gc_counter}};
  $dst->{gc_dont} = $src->{gc_dont};
  return $dst->SUPER::assign($src);
}

# $fs = $fs->remove_unreachable();
# $fs = $fs->remove_unreachable($node);
# ALIASES: remove_dangling(), trim()
*remove_dangling = \&remove_unreachable;
*trim = \&remove_unreachable;
*remove_unreachable_nodes = \&remove_unreachable;
*do_gc = \&remove_unreachable;
sub remove_unreachable {
  my $fs = shift;
  ${$fs->{gc_counter}} = 0 unless (@_);
  return $fs->SUPER::remove_unreachable(@_);
}

###############################################################
# Other gc hooks
###############################################################

# $fs->remove_nodes(@nodeids)
# $fs->remove_node(@nodeids)
# $fs->remove(@nodeids)
#*remove_node = \&remove_nodes;
#*remove = \&remove_nodes;
#sub remove_nodes {
#  my $fs = shift;
#  my $rv = $fs->SUPER::remove_nodes(@_);
#  $fs->_check_gc unless ($fs->{gc_dont});
#  return $rv;
#}

# $fs1_or_undef = $fs1->_unify($fs2);
sub _unify {
  my $fs = shift;
  $fs->{gc_dont} = 1;              # don't gc during unification
  my $rv = $fs->SUPER::_unify(@_); # ... actually do the unify
  $fs->{gc_dont} = 0;              # now we can gc again
  $fs->_check_gc() if ($rv);       # ... and we might need to
  return $rv;
}

1;
__END__

# -*- Mode: Perl -*-
=pod

=head1 NAME

Data::FStructure::Safe --
Feature-structures with automated garbage-collection.

=head1 SYNOPSIS

  use Data::FStructure::Safe;

  $fs = Data::FStructure::Safe->new(); # new FStructure

  $fs->root($fs->add_node);           # set the root of the FStructure
  $fs->pvalue([qw(syn cat)], 'verb'); # ... or any other FStructure method

  #
  # Low-Level Methods
  #
  $n = $fs->_gc_interval();           # get garbage-collection interval
  $i = $fs->_gc_counter();            # get internal operation-counter

  $fs->_gc_interval($n);              # set garbage-collection interval
  $fs->_gc_counter($i);               # set internal operation-counter

  $fs->_check_gc();                   # do gc if required


=head1 REQUIREMENTS

Data::FStructure::Base.

=head1 DESCRIPTION

C<Data::FStructure::Safe> is a subclass of C<Data::FStructure::Base>
which keeps track of the number of destructive operations
which have been performed on the underlying FStructure object,
and calls the garbage-collecting method C<remove_unreachable()>
after a user-specified threshold has been exceeded.

=head1 PACKAGE VARIABLES

The package-variables $GC_INTERVAL and $GC_VERBOSE
may be imported by specifying the ':gcvars' flag
to the C<use> directive.

=over 4

=item * C<$Data::FStructure::Safe::GC_INTERVAL>

This variable defines the default number of operations
that newly created Data::FStructure::Safe objects will allow
before initiating the garbage-collection routine.

=item * C<$Data::FStructure::Safe::GC_VERBOSE>

If this variable is set to a true value, a message will
be printed to STDERR every time garbage-collection
is initiated.  The default value is 0 (false).

=back

=head2 METHODS

=over 4

=item * C<new(\%args)>

Returns a new Data::FStructure::Safe object.  In addition
to the keyword arguments accepted by the C<Data::FStructure::Base>
method C<new()>, this method recognizes the following keys in C<\%args>:

=over 4

=item * C<< gc_interval =E<gt> $n >>

Specify the number of destructive operations that should
be allowed for the new object before garbage-collection
is initiated.  Default value is $GC_INTERVAL.

=item * C<< gc_counter =E<gt> $i >>

Specify the number of destructive operations that the
newly created object should consider it has already
undergone.  Default value is 0.

=item * C<< gc_dont =E<gt> $bool >>

Setting this flag to a true value will cause a
C<Data::FStructure::Safe> object to refrain from
garbage collecting.  This flag is set during
the unification process, when nodes and transitions
are added in apparently random order, and cleared
again on completion of the unification.

=back

=back

=head2 Low-Level Methods

=over 4

=item * C<_gc_interval()>, C<_gc_interval($n)>

Get/set the number of operations that should
be allowed before garbage-collection is initiated.
Setting this value may cause garbage-collection to be
initiated.

=item * C<_gc_counter()>, C<_gc_counter($n)>

Get/set the number of destructive operations already
undergone.  This value is maintained internally by
all C<Data::FStructure::Safe> objects, and probably
should never need to be called by the user.

=item * C<_check_gc()>

Checks whether the number of destructive operations already
undergone (as reported by C<_gc_counter()>) exceeds the
garbage-collection threshold (as reported by C<_gc_interval()>),
and initiates garbage-collection (by calling
C<remove_unreachable()>) if this is the case.
Increments C<_gc_counter> as a side effect.

=back

=head1 BUGS

See the section on "Dangling Nodes" in L<Data::FStructure::Base>
for details on garbage collection in
FStructure objects and possible future changes.

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<gt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2001, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).
Data::FStructure(3pm).
Data::FStructure::Base(3pm).
